# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
import torch
import roma
import unittest
from test import utils

class TestProcrustesDerivatives(unittest.TestCase):
    def test_derivatives(self):
        # Can only be done using float64
        device = torch.device(0) if torch.cuda.is_available() else None
        dtype=torch.float64
        batch_size = 10
        torch.manual_seed(666)
        d = 3

        M = torch.randn(batch_size, d, d, dtype=dtype, device=device)
        # Check derivatives
        eps = 1e-7
        eps2 = 1e-4
        for func in (lambda x : roma.procrustes(x),
                    lambda x : roma.special_procrustes(x),
                    lambda x: roma.procrustes_naive(x),
                    lambda x: roma.special_procrustes_naive(x),
                    lambda x: roma.procrustes(x, return_singular_values=True)[1],
                    lambda x: roma.special_procrustes(x, return_singular_values=True)[1],
                    lambda x: roma.procrustes_naive(x, return_singular_values=True)[1],
                    lambda x: roma.special_procrustes_naive(x, return_singular_values=True)[1],
                    ):
            # Numerical gradient
            num = utils.numerical_jacobian(func, M, eps)
            auto = utils.automatic_jacobian(func, M)
            self.assertTrue(utils.is_close(num, auto, eps2=eps2))

    def _test_convergence(self, random_initialization, regularization=0.0):
        """
        Try to solve an optimization problem using Special Procrustes on SO(3)
        """
        device=torch.device(0) if torch.cuda.is_available() else None
        torch.manual_seed(666)
        b = 5
        d = 3
        # Remark: does not converge to the true minimum  without enforcing rotation.
        force_rotation = True
        if random_initialization:
            # Random initialization
            M = torch.randn((b, d, d), requires_grad=True, device=device)
        else:
            # Harder degenerated case to test numerical stability
            M = torch.zeros((b, d, d), requires_grad=True, device=device)
            # If given a zero matrix as input, gradients are equal to 0 and nothing happens.
            M.data[:,0,0] = 1

        Rtarget = roma.random_rotmat(b, device=device)
        assert roma.is_rotation_matrix(Rtarget, 1e-5)

        optimizer = torch.optim.Adam([M], lr=0.1)

        # display_period = 100
        # print(f"Regularization: {regularization}")
        for iteration in range(2000):
            optimizer.zero_grad()

            R = roma.procrustes(M, force_rotation=force_rotation, regularization=regularization)
            if force_rotation:
                assert roma.is_rotation_matrix(R, 1e-5)
            else:
                assert roma.is_orthonormal_matrix(R, 1e-5)
            loss = torch.nn.functional.mse_loss(R, Rtarget)
            with torch.no_grad():
                unnormalized_loss = torch.nn.functional.mse_loss(R, M)
            # if iteration % display_period == display_period-1:
            #     print(f"{iteration}: loss {loss.item()} -- unnormalized_loss {unnormalized_loss.item()}")
            
            loss.backward()
            optimizer.step()
        self.assertLess(loss.item(),  1e-7)
        if regularization > 0:
            # M should be roughly equal to R after optimization due to the regularization
            self.assertLess(unnormalized_loss.item(),  1e-4)

    def test_convergence_random_initialization(self):
        self._test_convergence(True, 0.0)
        self._test_convergence(True, 1e-4)

    def test_convergence_degenerated_initialization(self):
        self._test_convergence(False)    
        self._test_convergence(False, 1e-4)

if __name__ == '__main__':
    unittest.main()


# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
import torch
import unittest
from test.utils import *

# Test of test utils
class TestDerivatives(unittest.TestCase):
    def test_derivatives(self):
        # Test the derivatives
        x = torch.randn((2,3,4), dtype=torch.float64)
        func = torch.tanh
        auto = automatic_jacobian(func, x)
        num = numerical_jacobian(func, x, 1e-10)
        self.assertTrue(is_close(auto, num, 1e-6))

        func = lambda x : torch.sum(torch.tanh(x))
        auto = automatic_jacobian(func, x)
        num = numerical_jacobian(func, x, 1e-10)
        self.assertTrue(is_close(auto, num, 1e-6))
        
        x = torch.randn((2, 3, 3), dtype=torch.float64)
        func = lambda x: torch.sum(x[0] @ x[1], axis=1)
        auto = automatic_jacobian(func, x)
        num = numerical_jacobian(func, x, 1e-10)
        self.assertTrue(is_close(auto, num, 1e-6))

if __name__ == "__main__":
    unittest.main()        


# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
import unittest
import torch
import roma
import numpy as np
from test.utils import is_close

def is_close(A, B, eps1 = 1e-5, eps2 = 1e-5):
    return torch.norm(A - B) / (torch.norm(torch.abs(A) + torch.abs(B)) + eps1) < eps2

device = torch.device(0) if torch.cuda.is_available() else torch.device('cpu')

class TestMappings(unittest.TestCase):
    def test_orthonormal(self):
        for dtype in (torch.float32, torch.float64):
            M = torch.eye(3, dtype=dtype, device=device).expand(10, 3, 3).contiguous()
            self.assertTrue(roma.is_orthonormal_matrix(M))
            M[:,:,-1] *= -1
            self.assertTrue(roma.is_orthonormal_matrix(M))
            torch.manual_seed(666)
            M = torch.randn((1,3,3))
            self.assertFalse(roma.is_orthonormal_matrix(M))
        
    def test_rotation(self):
        for dtype in (torch.float32, torch.float64):
            M = torch.eye(3, dtype=dtype, device=device).expand(10, 3, 3).contiguous()
            self.assertTrue(roma.is_rotation_matrix(M))
            M[:,:,-1] *= -1
            self.assertFalse(roma.is_rotation_matrix(M))
            torch.manual_seed(666)
            M = torch.randn((1,3,3))
            self.assertFalse(roma.is_rotation_matrix(M))
        
    def test_procrustes(self):
        torch.manual_seed(666)
        for dtype in (torch.float32, torch.float64):
            for i in range(10):
                M = torch.randn((10,3,3), dtype=dtype, device=device)
                R = roma.procrustes(M)
                self.assertTrue(roma.is_orthonormal_matrix(R, 1e-5))
                Rbis = roma.procrustes(R)
                self.assertTrue(is_close(R, Rbis))
                # Ensure consistency
                Rter = roma.procrustes_naive(M)
                self.assertTrue(is_close(R, Rter))
            
    def test_special_procrustes(self):
        torch.manual_seed(666)
        for dtype in (torch.float32, torch.float64):
            for i in range(10):
                M = torch.randn((10,3,3), dtype=dtype, device=device)
                R = roma.special_procrustes(M)
                self.assertTrue(roma.is_rotation_matrix(R, 1e-5))
                Rbis = roma.special_procrustes(R)
                self.assertTrue(is_close(R, Rbis))
                # Ensure consistency
                Rter = roma.special_procrustes_naive(M)
                self.assertTrue(is_close(R, Rter))
            
    def test_special_gramschmidt(self):
        torch.manual_seed(666)
        for dtype in (torch.float32, torch.float64):
            M = torch.randn((100,3,2), dtype=dtype, device=device)
            R = roma.special_gramschmidt(M)
            self.assertTrue(roma.is_rotation_matrix(R, 1e-5))
            Rbis = roma.special_gramschmidt(R)
            self.assertTrue(is_close(R, Rbis))
        
    def test_rotvec_unitquat(self):
        torch.manual_seed(666)
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            x = 10 * torch.randn((batch_size, 3), dtype=dtype, device=device)
            #Forward mapping.
            q = roma.rotvec_to_unitquat(x)
            # Ensure that the output is a valid unit quaternion.
            self.assertEqual(q.shape, (batch_size, 4))
            self.assertTrue(torch.all(torch.abs(torch.norm(q, dim=-1) - 1) < 1e-6))

            # Backward mapping
            xbis = roma.unitquat_to_rotvec(q)
            qbis = roma.rotvec_to_unitquat(xbis)
            # Ensure cyclic consistency of the mappings
            self.assertTrue(torch.all(torch.min(torch.norm(qbis - q, dim=-1), torch.norm(qbis + q, dim=-1)) < 1e-6))
            xter = roma.unitquat_to_rotvec(qbis)
            self.assertTrue(is_close(xbis, xter))


    def test_unitquat_to_rotvec(self):
        torch.manual_seed(666)
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            q = roma.random_unitquat(batch_size, dtype)
            xp = roma.unitquat_to_rotvec(q, shortest_arc=True)
            xm = roma.unitquat_to_rotvec(-q, shortest_arc=True)
            self.assertEqual(xp.shape, (batch_size, 3))
            self.assertTrue(is_close(xp, xm))
            self.assertTrue(torch.all(torch.norm(xp, dim=-1) <= np.pi))

            # Mappings from q and -q should give different results in general when shortest_arc==False
            xpn = roma.unitquat_to_rotvec(q, shortest_arc=False)
            xmn = roma.unitquat_to_rotvec(-q, shortest_arc=False)
            self.assertFalse(any([is_close(xpn[i], xmn[i]) for i in range(batch_size)]))

            # However, the result should be similar when shortest_arc==True
            xpn = roma.unitquat_to_rotvec(q, shortest_arc=True)
            xmn = roma.unitquat_to_rotvec(-q, shortest_arc=True)
            self.assertTrue(is_close(xpn, xmn))

            # Ensure cyclic consistency of the mappings in any case
            for x in xp, xm, xpn, xmn:
                qbis = roma.rotvec_to_unitquat(x)
                self.assertTrue(torch.all(torch.min(torch.norm(qbis - q, dim=-1), torch.norm(qbis + q, dim=-1)) < 5e-6))

    def test_rotvec_rotmat(self):
        torch.manual_seed(666)
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            # Perform the test for large and small angles
            for scale in (10.0, 1e-7):
                x = scale * torch.randn((batch_size, 3), dtype=dtype, device=device)
                #Forward mapping
                R = roma.rotvec_to_rotmat(x)
                self.assertTrue(roma.is_rotation_matrix(R, 1e-5))
                # Backward mapping
                xbis = roma.rotmat_to_rotvec(R)
                Rbis = roma.rotvec_to_rotmat(x)
                self.assertTrue(is_close(R, Rbis))
                xter = roma.rotmat_to_rotvec(Rbis)
                self.assertTrue(is_close(xbis, xter))        
        
    def test_unitquat_rotmat(self):
        torch.manual_seed(666)
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            q = roma.random_unitquat(batch_size, dtype=dtype, device=device)
            # Forward
            R = roma.unitquat_to_rotmat(q)
            self.assertTrue(roma.is_rotation_matrix(R, 1e-5))
            # Backward
            qbis = roma.rotmat_to_unitquat(R)
            self.assertTrue(torch.all(torch.min(torch.norm(qbis - q, dim=-1), torch.norm(qbis + q, dim=-1)) < 1e-6))
      
    def test_symmatrix_to_unitquat(self):
        torch.manual_seed(668)
        batch_size = 100
        # Remark: the eigenvalue decomposition may fail using float32 or depending on the random seed,
        # due to conditionning issues.
        for dtype in (torch.float32, torch.float64):
            x = torch.randn((batch_size, 10), dtype=dtype, device=device)
            q = roma.symmatrixvec_to_unitquat(x)
            # Ensure that the output is a unit quaternion.
            self.assertEqual(q.shape, (batch_size, 4))
            self.assertTrue(torch.all(torch.abs(torch.norm(q, dim=-1) - 1) < 1e-6))

            # Same test using explicitly a symmetric matrix.
            A = torch.randn((batch_size, 4, 4), dtype=dtype, device=device)
            A = A + A.permute(0, 2, 1)
            q = roma.symmatrix_to_projective_point(A)
            # Ensure that the output is a unit quaternion.
            self.assertEqual(q.shape, (batch_size, 4))
            self.assertTrue(torch.all(torch.abs(torch.norm(q, dim=-1) - 1) < 1e-6))

    def test_rotvec_rotmat_nan_issues(self):
        # Check that casting rotation vectors to rotation matrices does not lead to non finite values with 0 angles.
        rotvec = torch.zeros(3, requires_grad=True)
        R = roma.rotvec_to_rotmat(rotvec)
        loss = torch.sum(R)
        loss.backward()
        self.assertTrue(torch.all(torch.isfinite(loss)))
        self.assertTrue(torch.all(torch.isfinite(rotvec.grad)))

    def test_rotvec_unitquat_nan_issues(self):
        """
        Check that casting using rotation vectors does not lead to non finite values with 0 angles.
        """
        rotvec = torch.zeros(3, requires_grad=True)
        q = roma.rotvec_to_unitquat(rotvec)
        loss = torch.sum(q)
        loss.backward()
        self.assertTrue(torch.all(torch.isfinite(loss)))
        self.assertTrue(torch.all(torch.isfinite(rotvec.grad)))

    def test_quat_conventions(self):
        for batch_shape in [(), (10,), (23,5)]:
            quat_xyzw = torch.randn(batch_shape + (4,))
            quat_wxyz = roma.mappings.quat_xyzw_to_wxyz(quat_xyzw)
            self.assertTrue(quat_xyzw.shape == quat_wxyz.shape)
            quat_xyzw_bis = roma.mappings.quat_wxyz_to_xyzw(quat_wxyz)
            self.assertTrue(torch.all(quat_xyzw == quat_xyzw_bis))


if __name__ == "__main__":
    unittest.main()


# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
"""
Various mappings between different rotation representations.
"""

import torch
import roma.internal

class _ProcrustesManualDerivatives(torch.autograd.Function):
    @staticmethod
    def forward(ctx, M, force_rotation, regularization, gradient_eps):
        assert (M.dim() == 3 and M.shape[1] == M.shape[2]), "Input should be a BxDxD batch of matrices."
        # Singular values of D are sorted in descending order
        U, D, V = roma.internal.svd(M)
        if force_rotation:
            # We flip the smallest singular value to ensure getting a rotation matrix
            with torch.no_grad():
                flip = (torch.det(U) * torch.det(V) < 0)            
            # in-place modifications of variables not used afterwards.
            DS = D
            DS[flip,-1] *= -1
            del D
            US = U
            US[flip,:,-1] *= -1
            del U
        else:
            flip = None
            DS = D
            US = U
        R = US @ V.transpose(-1, -2)
        # Store data for backprop:
        ctx.save_for_backward(US, DS, V, M, R)
        ctx.gradient_eps = gradient_eps
        ctx.regularization = regularization
        return R, DS

    @staticmethod
    def backward(ctx, grad_R, grad_DS):
        US, DS, V, M, R = ctx.saved_tensors
        gradient_eps = ctx.gradient_eps

        USik_Vjl = torch.einsum('bik,bjl -> bklij', US, V)
        USil_Vjk = USik_Vjl.transpose(1,2)
        DSl = DS[:,None,:,None,None]
        DSk = DS[:,:,None,None,None]
        Omega_klij = (USik_Vjl - USil_Vjk) * roma.internal._pseudo_inverse(DSk + DSl, gradient_eps)
        # Note: this intermediary matrix may require lots of memory for large dimensional cases.
        # Diagonal k==l should always be 0 thanks to the clamping of the pseudo-inverse.
        
        grad_M = torch.einsum('bnm, bnk, bklij, bml -> bij', grad_R, US, Omega_klij, V)
        # Gradient contribution from singular values
        grad_M += (US * grad_DS[:,None,:]) @ V.transpose(-1, -2)
        if ctx.regularization != 0.0:
            # Add a regularization term in the direction of the orthonormalized output.
            grad_M += ctx.regularization * (M - R)
        return grad_M, None, None, None

def procrustes(M, force_rotation=False, regularization=0.0, gradient_eps=1e-5, return_singular_values : bool = False):
    """ 
    Returns the orthonormal matrix :math:`R` minimizing Frobenius norm :math:`\| M - R \|_F`.

    Args:
        M (...xNxN tensor): batch of square matrices.
        force_rotation (bool): if True, forces the output to be a rotation matrix.
        regularization (float >= 0): weight of a regularization term added to the gradient.
            Using this regularization is equivalent to adding a term :math:`regularization * \| M - R \|_F^2` to the training loss function.
        gradient_eps (float > 0): small value used to enforce numerical stability during backpropagation.
    Returns:
        batch of orthonormal matrices (...xNxN tensor) and optional singular values.
        For advanced users, singular values of the SVD decomposition with sign flipping (... tensor) can optionally be returned by setting the argument :code:`return_singular_values` to :code:`True`.
    """
    M, batch_shape = roma.internal.flatten_batch_dims(M, -3)
    R, DS = _ProcrustesManualDerivatives.apply(M, force_rotation, regularization, gradient_eps)
    R = roma.internal.unflatten_batch_dims(R, batch_shape)
    if not return_singular_values:
        return R
    else:
        DS = roma.internal.unflatten_batch_dims(DS, batch_shape)
        return R, DS

def special_procrustes(M, regularization=0.0, gradient_eps=1e-5, return_singular_values : bool = False):
    """
    Returns the rotation matrix :math:`R` minimizing Frobenius norm :math:`\| M - R \|_F`.

    Args:
        M (...xNxN tensor): batch of square matrices.
        regularization (float >= 0): weight of a regularization term added to the gradient.
            Using this regularization is equivalent to adding a term :math:`regularization * \| M - R \|_F^2` to the training loss function.
        gradient_eps (float > 0): small value used to enforce numerical stability during backpropagation.
    Returns:
        batch of rotation matrices (...xNxN tensor).
        For advanced users, singular values of the SVD decomposition with sign flipping (... tensor) can optionally be returned by setting the argument :code:`return_singular_values` to :code:`True`.

    """
    return procrustes(M, True, regularization, gradient_eps, return_singular_values)

def procrustes_naive(M, force_rotation : bool = False, return_singular_values : bool = False):
    """
    Implementation of :func:`~roma.mappings.procrustes` relying on default backward pass of autograd and SVD decomposition.
    Could be slightly less stable than :func:`~roma.mappings.procrustes`.
    """
    M, batch_shape = roma.internal.flatten_batch_dims(M, -3)
    assert (M.dim() == 3 and M.shape[1] == M.shape[2]), "Input should be a BxDxD batch of matrices."
    U, D, V = roma.internal.svd(M)
    # D is sorted in descending order
    SVt = V.transpose(-1,-2)
    if force_rotation:
        # We flip the smallest singular value to ensure getting a rotation matrix
        with torch.no_grad():
            flip = (torch.det(U) * torch.det(V) < 0)
        if torch.is_grad_enabled():
            # This is needed to avoid a runtime error "one of the variables needed for gradient computation has been modified by an inplace operation"
            SVt = SVt.clone()
        SVt[flip,-1,:] *= -1
    R = U @ SVt
    R = roma.internal.unflatten_batch_dims(R, batch_shape)
    if not return_singular_values:
        return R
    else:
        DS = D.clone()
        if force_rotation:
            DS[flip, -1] *= -1
        del D
        return R, DS


def special_procrustes_naive(M, return_singular_values : bool = False):
    """
    Implementation of :func:`~roma.mappings.special_procrustes` relying on default backward pass of autograd and SVD decomposition.
    Could be slightly less stable than :func:`~roma.mappings.special_procrustes`.
    """
    return procrustes_naive(M, force_rotation=True, return_singular_values=return_singular_values)

def special_gramschmidt(M, epsilon=0):
    """
    Returns the 3x3 rotation matrix obtained by Gram-Schmidt orthonormalization of two 3D input vectors (first two columns of input matrix M).

    Args:
        M (...x3xN tensor): batch of 3xN matrices, with N >= 2.
            Only the first two columns of the matrices are used for orthonormalization.
        epsilon (float >= 0): optional clamping value to avoid returning *Not-a-Number* values in case of ill-defined input.
    Returns:
        batch of rotation matrices (...x3x3 tensor).
    Warning:
        In case of ill-defined input (colinear input column vectors), the output will not be a rotation matrix.
    """
    M, batch_shape = roma.internal.flatten_batch_dims(M, -3)
    assert(M.dim() == 3)
    x = M[:,:,0]
    y = M[:,:,1]
    x = x / torch.clamp_min(torch.norm(x, dim=-1, keepdim=True), epsilon)
    y = y - torch.sum(x*y, dim=-1, keepdim=True) * x
    y = y / torch.clamp_min(torch.norm(y, dim=-1, keepdim=True), epsilon)
    z = torch.cross(x,y, dim=-1)    
    R = torch.stack((x, y, z), dim=-1)
    return roma.internal.unflatten_batch_dims(R, batch_shape)

def symmatrix_to_projective_point(A):
    """
    Converts a DxD symmetric matrix A into a projective point represented by a unit vector :math:`q` minimizing :math:`q^T A q`.
    Only the lower part of the matrix is considered in practice.

    Args:
        A (...xDxD tensor): batch of symmetric matrices. Only the lower triangular part is considered.
    Returns:
        batch of unit vectors :math:`q` (...xD tensor).
    Reference:
        V. Peretroukhin, M. Giamou, D. M. Rosen, W. N. Greene, N. Roy, and J. Kelly, “A Smooth Representation of Belief over SO(3) for Deep Rotation Learning with Uncertainty,” 2020, doi: 10.15607/RSS.2020.XVI.007.

    Warning:
        - This mapping is unstable when the smallest eigenvalue of A has a multiplicity strictly greater than 1.
        - The eigenvalue decomposition may fail, in particular when using single precision numbers.
        - Current implementation is rather slow due to the implementation of ``torch.symeig``.
          The CuSolver library provides a faster eigenvalue decomposition alternative, but results where found to be unreliable.
    """
    A, batch_shape = roma.internal.flatten_batch_dims(A, end_dim=-3)
    B, D1, D2 = A.shape
    assert (D1,D2) == (4,4), "Input should be a symmetric Bx4x4 matrix."
    eigenvalues, eigenvectors = roma.internal.symeig_lower(A)
    # Eigenvalues are sorted in ascending order
    q = eigenvectors[:,:,0]
    return roma.internal.unflatten_batch_dims(q, batch_shape)

def symmatrixvec_to_unitquat(x):
    """
    Converts a 10D vector into a unit quaternion representation.
    Based on :func:`~roma.mappings.symmatrix_to_projective_point`.
    
    Args:
        x (...x10 tensor): batch of 10D vectors.
    Returns:
        batch of unit quaternions (...x4 tensor, XYZW convention).
    Reference:
        V. Peretroukhin, M. Giamou, D. M. Rosen, W. N. Greene, N. Roy, and J. Kelly, “A Smooth Representation of Belief over SO(3) for Deep Rotation Learning with Uncertainty,” 2020, doi: 10.15607/RSS.2020.XVI.007.
    """
    x, batch_shape = roma.internal.flatten_batch_dims(x, end_dim=-2)
    batch_size, D = x.shape
    assert(D) == 10, "Input should be a Bx10 tensor."
    A = torch.empty((batch_size, 4, 4), dtype=x.dtype, device=x.device)
    # Fill only the lower triangular part of the matrix.
    A[:, :,0] = x[:,0:4]
    A[:,1:,1] = x[:,4:7]
    A[:,2:,2] = x[:,7:9]
    A[:, 3,3] = x[:,9]
    return roma.internal.unflatten_batch_dims(symmatrix_to_projective_point(A), batch_shape)    

def rotvec_to_unitquat(rotvec):
    """
    Converts rotation vector into unit quaternion representation.

    Args:
        rotvec (...x3 tensor): batch of rotation vectors.
    Returns:
        batch of unit quaternions (...x4 tensor, XYZW convention).
    """
    rotvec, batch_shape = roma.internal.flatten_batch_dims(rotvec, end_dim=-2)
    num_rotations, D = rotvec.shape
    assert D == 3, "Input should be a Bx3 tensor."

    # Adapted from SciPy:
    # https://github.com/scipy/scipy/blob/adc4f4f7bab120ccfab9383aba272954a0a12fb0/scipy/spatial/transform/rotation.py#L621
    
    norms = torch.norm(rotvec, dim=-1)
    small_angle = (norms <= 1e-3)
    large_angle = ~small_angle

    scale = torch.empty((num_rotations,), device=rotvec.device, dtype=rotvec.dtype)
    scale[small_angle] = (0.5 - norms[small_angle] ** 2 / 48 +
                          norms[small_angle] ** 4 / 3840)
    scale[large_angle] = (torch.sin(norms[large_angle] / 2) /
                          norms[large_angle])

    quat = torch.empty((num_rotations, 4), device=rotvec.device, dtype=rotvec.dtype)
    quat[:, :3] = scale[:, None] * rotvec
    quat[:, 3] = torch.cos(norms / 2)
    return roma.internal.unflatten_batch_dims(quat, batch_shape)

def unitquat_to_rotvec(quat, shortest_arc=True):
    """
    Converts unit quaternion into rotation vector representation.

    Based on the representation of a rotation of angle :math:`{\\theta}` and unit axis :math:`(x,y,z)` 
    by the unit quaternions :math:`\pm [\sin({\\theta} / 2) (x i + y j + z k) + \cos({\\theta} / 2)]`.

    Args:
        quat (...x4 tensor, XYZW convention): batch of unit quaternions.
            No normalization is applied before computation.
        shortest_arc (bool): if True, the function returns the smallest rotation vectors corresponding 
            to the input 3D rotations, i.e. rotation vectors with a norm smaller than :math:`\pi`.
            If False, the function may return rotation vectors of norm larger than :math:`\pi`, depending on the sign of the input quaternions.
    Returns:
        batch of rotation vectors (...x3 tensor).
    Note:
        Behavior is undefined for inputs ``quat=torch.as_tensor([0.0, 0.0, 0.0, -1.0])`` and ``shortest_arc=False``, 
        as any rotation vector of angle :math:`2 \pi` could be a valid representation in such case.
    """
    quat, batch_shape = roma.internal.flatten_batch_dims(quat, end_dim=-2)
    # We perform a copy to support auto-differentiation.
    quat = quat.clone()
    # Adapted from SciPy:
    # https://github.com/scipy/scipy/blob/adc4f4f7bab120ccfab9383aba272954a0a12fb0/scipy/spatial/transform/rotation.py#L1006-L1073
    if shortest_arc:
        # Enforce w > 0 to ensure 0 <= angle <= pi.
        # (Otherwise angle can be arbitrary within ]-2pi, 2pi]).
        quat[quat[:, 3] < 0] *= -1
    half_angle = torch.atan2(torch.norm(quat[:, :3], dim=1), quat[:, 3])
    angle = 2 * half_angle
    small_angle = (torch.abs(angle) <= 1e-3)
    large_angle = ~small_angle

    num_rotations = len(quat)
    scale = torch.empty(num_rotations, dtype=quat.dtype, device=quat.device)
    scale[small_angle] = (2 + angle[small_angle] ** 2 / 12 +
                          7 * angle[small_angle] ** 4 / 2880)
    scale[large_angle] = (angle[large_angle] /
                          torch.sin(half_angle[large_angle]))

    rotvec = scale[:, None] * quat[:, :3]
    return roma.internal.unflatten_batch_dims(rotvec, batch_shape)

def unitquat_to_rotmat(quat):
    """
    Converts unit quaternion into rotation matrix representation.

    Args:
        quat (...x4 tensor, XYZW convention): batch of unit quaternions.
            No normalization is applied before computation.
    Returns:
        batch of rotation matrices (...x3x3 tensor).
    """
    # Adapted from SciPy:
    # https://github.com/scipy/scipy/blob/adc4f4f7bab120ccfab9383aba272954a0a12fb0/scipy/spatial/transform/rotation.py#L912
    x = quat[..., 0]
    y = quat[..., 1]
    z = quat[..., 2]
    w = quat[..., 3]

    x2 = x * x
    y2 = y * y
    z2 = z * z
    w2 = w * w

    xy = x * y
    zw = z * w
    xz = x * z
    yw = y * w
    yz = y * z
    xw = x * w

    matrix = torch.empty(quat.shape[:-1] + (3, 3), dtype=quat.dtype, device=quat.device)

    matrix[..., 0, 0] = x2 - y2 - z2 + w2
    matrix[..., 1, 0] = 2 * (xy + zw)
    matrix[..., 2, 0] = 2 * (xz - yw)

    matrix[..., 0, 1] = 2 * (xy - zw)
    matrix[..., 1, 1] = - x2 + y2 - z2 + w2
    matrix[..., 2, 1] = 2 * (yz + xw)

    matrix[..., 0, 2] = 2 * (xz + yw)
    matrix[..., 1, 2] = 2 * (yz - xw)
    matrix[..., 2, 2] = - x2 - y2 + z2 + w2
    return matrix

def rotmat_to_unitquat(R):
    """
    Converts rotation matrix to unit quaternion representation.

    Args:
        R (...x3x3 tensor): batch of rotation matrices.
    Returns:
        batch of unit quaternions (...x4 tensor, XYZW convention).
    """
    matrix, batch_shape = roma.internal.flatten_batch_dims(R, end_dim=-3)
    num_rotations, D1, D2 = matrix.shape
    assert((D1, D2) == (3,3)), "Input should be a Bx3x3 tensor."

    # Adapted from SciPy:
    # https://github.com/scipy/scipy/blob/7cb3d751756907238996502b92709dc45e1c6596/scipy/spatial/transform/rotation.py#L480

    decision_matrix = torch.empty((num_rotations, 4), dtype=matrix.dtype, device=matrix.device)
    decision_matrix[:, :3] = matrix.diagonal(dim1=1, dim2=2)
    decision_matrix[:, -1] = decision_matrix[:, :3].sum(axis=1)
    choices = decision_matrix.argmax(axis=1)

    quat = torch.empty((num_rotations, 4), dtype=matrix.dtype, device=matrix.device)

    ind = torch.nonzero(choices != 3, as_tuple=True)[0]
    i = choices[ind]
    j = (i + 1) % 3
    k = (j + 1) % 3

    quat[ind, i] = 1 - decision_matrix[ind, -1] + 2 * matrix[ind, i, i]
    quat[ind, j] = matrix[ind, j, i] + matrix[ind, i, j]
    quat[ind, k] = matrix[ind, k, i] + matrix[ind, i, k]
    quat[ind, 3] = matrix[ind, k, j] - matrix[ind, j, k]

    ind = torch.nonzero(choices == 3, as_tuple=True)[0]
    quat[ind, 0] = matrix[ind, 2, 1] - matrix[ind, 1, 2]
    quat[ind, 1] = matrix[ind, 0, 2] - matrix[ind, 2, 0]
    quat[ind, 2] = matrix[ind, 1, 0] - matrix[ind, 0, 1]
    quat[ind, 3] = 1 + decision_matrix[ind, -1]

    quat = quat / torch.norm(quat, dim=1)[:, None]
    return roma.internal.unflatten_batch_dims(quat, batch_shape)

def rotvec_to_rotmat(rotvec: torch.Tensor, epsilon=1e-6) -> torch.Tensor:
    """
    Converts rotation vector to rotation matrix representation.
    Conversion uses Rodrigues formula in general, and a first order approximation for small angles.
    
    Args:
        rotvec (...x3 tensor): batch of rotation vectors.
        epsilon (float): small angle threshold.
    Returns:
        batch of rotation matrices (...x3x3 tensor).        
    """
    rotvec, batch_shape = roma.internal.flatten_batch_dims(rotvec, end_dim=-2)
    batch_size, D = rotvec.shape
    assert(D == 3), "Input should be a Bx3 tensor."

    # Rotation angle
    theta = torch.norm(rotvec, dim=-1)
    is_angle_small = theta < epsilon
    
    # Rodrigues formula for angles that are not small.
    # Note: we use clamping to avoid non finite values for small angles
    # (torch.where produces nan gradients in such case).
    axis = rotvec / torch.clamp_min(theta[...,None], epsilon)
    kx, ky, kz = axis[:,0], axis[:,1], axis[:,2]
    sin_theta = torch.sin(theta)
    cos_theta = torch.cos(theta)
    one_minus_cos_theta = 1 - cos_theta
    xs = kx*sin_theta
    ys = ky*sin_theta
    zs = kz*sin_theta
    xyc = kx*ky*one_minus_cos_theta
    xzc = kx*kz*one_minus_cos_theta
    yzc = ky*kz*one_minus_cos_theta
    xxc = kx**2*one_minus_cos_theta
    yyc = ky**2*one_minus_cos_theta
    zzc = kz**2*one_minus_cos_theta
    R_rodrigues = torch.stack([1 - yyc - zzc, xyc - zs, xzc + ys,
                     xyc + zs, 1 - xxc - zzc, -xs + yzc,
                     xzc - ys, xs + yzc, 1 -xxc - yyc], dim=-1).reshape(-1, 3, 3)

    # For small angles, use a first order approximation
    xs, ys, zs = rotvec[:,0], rotvec[:,1], rotvec[:,2]
    one = torch.ones_like(xs)
    R_first_order = torch.stack([one, -zs, ys,
                                 zs, one, -xs,
                                 -ys, xs, one], dim=-1).reshape(-1, 3, 3)
    # Select the appropriate expression
    R = torch.where(is_angle_small[:,None,None], R_first_order, R_rodrigues)
    return roma.internal.unflatten_batch_dims(R, batch_shape)
    
def rotmat_to_rotvec(R):
    """
    Converts rotation matrix to rotation vector representation.

    Args:
        R (...x3x3 tensor): batch of rotation matrices.
    Returns:
        batch of rotation vectors (...x3 tensor).
    """
    q = rotmat_to_unitquat(R)
    return unitquat_to_rotvec(q)

def quat_xyzw_to_wxyz(xyzw):
    """
    Convert quaternion from XYZW to WXYZ convention.

    Args:
        xyzw (...x4 tensor, XYZW convention): batch of quaternions.
    Returns:
        batch of quaternions (...x4 tensor, WXYZ convention).
    """
    assert xyzw.shape[-1] == 4
    return torch.cat((xyzw[...,-1,None], xyzw[...,:-1]), dim=-1)

def quat_wxyz_to_xyzw(wxyz):
    """
    Convert quaternion from WXYZ to XYZW convention.

    Args:
        wxyz (...x4 tensor, WXYZ convention): batch of quaternions.
    Returns:
        batch of quaternions (...x4 tensor, XYZW convention).
    """
    assert wxyz.shape[-1] == 4
    return torch.cat((wxyz[...,1:], wxyz[...,0,None]), dim=-1)



# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
"""
Various utility functions related to rotation representations.
"""

import torch
import numpy as np
import roma.internal
import roma.mappings

def is_torch_batch_svd_available() -> bool:
    """
    Returns True if the module 'torch_batch_svd' has been loaded. Returns False otherwise.
    """
    return roma.internal._IS_TORCH_BATCH_SVD_AVAILABLE

def is_orthonormal_matrix(R, epsilon=1e-7):
    """
    Test if matrices are orthonormal.

    Args:
        R (...xDxD tensor): batch of square matrices.
        epsilon: tolerance threshold.
    Returns:
        boolean tensor (shape ...).

    """
    R, batch_shape = roma.internal.flatten_batch_dims(R, end_dim=-3)
    assert (R.dim() == 3), "Input should be a BxDxD batch of matrices."
    B, D, D1 = R.shape
    assert D == D1, "Input should be a BxDxD batch of matrices."
    errors = torch.norm(R @ R.transpose(-1, -2) - torch.eye(D, device=R.device, dtype=R.dtype), dim=[-2,-1])
    return torch.all(errors < epsilon)
    
def is_rotation_matrix(R, epsilon=1e-7):
    """
    Test if matrices are rotation matrices.

    Args:
        R (...xDxD tensor): batch of square matrices.
        epsilon: tolerance threshold.
    Returns:
        boolean tensor (shape ...).
    """
    if not is_orthonormal_matrix(R, epsilon):
        return False
    return torch.all(torch.det(R) > 0)

def random_unitquat(size = tuple(), dtype=torch.float, device=None):
    """
    Generates a batch of random unit quaternions, uniformly sampled according to the usual quaternion metric.

    Args:
        size (tuple or int): batch size. Use for example ``tuple()`` to generate a single element, and ``(5,2)`` to generate a 5x2 batch.
    Returns:
        batch of unit quaternions (size x 4 tensor).

    Reference:
        K. Shoemake, “Uniform Random Rotations”, in Graphics Gems III (IBM Version), Elsevier, 1992, pp. 124–132. doi: 10.1016/B978-0-08-050755-2.50036-1.

    """
    if type(size) == int:
        size = (size,)

    x0 = torch.rand(size, dtype=dtype, device=device)
    theta1 = (2.0 * np.pi) * torch.rand(size, dtype=dtype, device=device)
    theta2 = (2.0 * np.pi) * torch.rand(size, dtype=dtype, device=device)
    r1 = torch.sqrt(1.0 - x0)
    r2 = torch.sqrt(x0)
    return torch.stack((r1 * torch.sin(theta1), r1 * torch.cos(theta1), r2 * torch.sin(theta2), r2 * torch.cos(theta2)), dim=-1)

def random_rotmat(size  = tuple(), dtype=torch.float, device=None):
    """
    Generates a batch of random 3x3 rotation matrices, uniformly sampled according to the usual rotation metric.

    Args:
        size (tuple or int): batch size. Use for example ``tuple()`` to generate a single element, and ``(5,2)`` to generate a 5x2 batch.
    Returns:
        batch of rotation matrices (size x 3x3 tensor).
    """
    quat = random_unitquat(size, dtype=dtype, device=device)
    R = roma.mappings.unitquat_to_rotmat(quat)
    return R

def random_rotvec(size = tuple(), dtype=torch.float, device=None):
    """
    Generates a batch of random rotation vectors, uniformly sampled according to the usual rotation metric.

    Args:
        size (tuple or int): batch size. Use for example ``tuple()`` to generate a single element, and ``(5,2)`` to generate a 5x2 batch.
    Returns:
        batch of rotation vectors (size x 3 tensor).
    """
    quat = random_unitquat(size, dtype=dtype, device=device)
    return roma.mappings.unitquat_to_rotvec(quat)

def identity_quat(size = tuple(), dtype=torch.float, device=None):
    """
    Return a batch of identity unit quaternions.

    Args:
        size (tuple or int): batch size. Use for example ``tuple()`` to generate a single element, and ``(5,2)`` to generate a 5x2 batch.
    Returns:
        batch of identity quaternions (size x 4 tensor, XYZW convention).
    Note:
        All returned batch quaternions refer to the same memory location.
        Consider cloning the output tensor prior performing any in-place operations.
    """
    if type(size) == int:
        size = (size,)
    quat = torch.zeros(4, dtype=dtype, device=device)
    quat[...,-1] = 1.
    quat = quat.expand(list(size) + [-1])
    return quat

def rotmat_cosine_angle(R):
    """
    Returns the cosine angle of the input 3x3 rotation matrix R.
    Based on the equality :math:`Trace(R) = 1 + 2 cos(alpha)`.

    Args:
        R (...x3x3 tensor): batch of 3w3 rotation matrices.
    Returns:
        batch of cosine angles (... tensor).
    """
    assert R.shape[-2:] == (3,3), "Expecting a ...x3x3 batch of rotation matrices"
    return  0.5 * (R[...,0,0] + R[...,1,1] + R[...,2,2] - 1.0)

_ONE_OVER_2SQRT2 = 1.0 / (2 * np.sqrt(2))
def rotmat_geodesic_distance(R1, R2, clamping=1.0):
    """
    Returns the angular distance alpha between a pair of rotation matrices.
    Based on the equality :math:`|R_2 - R_1|_F = 2 \sqrt{2} sin(alpha/2)`.

    Args:
        R1, R2 (...x3x3 tensor): batch of 3x3 rotation matrices.
        clamping: clamping value applied to the input of :func:`torch.asin()`.
                Use 1.0 to ensure valid angular distances.
                Use a value strictly smaller than 1.0 to ensure finite gradients.
    Returns:
        batch of angles in radians (... tensor).
    """
    return 2.0 * torch.asin(torch.clamp_max(torch.norm(R2 - R1, dim=[-1, -2]) * _ONE_OVER_2SQRT2, clamping))

def rotmat_geodesic_distance_naive(R1, R2):
    """
    Returns the angular distance between a pair of rotation matrices.
    Based on :func:`~rotmat_cosine_angle` and less precise than :func:`~roma.utils.rotmat_geodesic_distance` for nearby rotations.
    
    Args:
        R1, R2 (...x3x3 tensor): batch of 3x3 rotation matrices.
    Returns:
        batch of angles in radians (... tensor).
    """
    R = R1.transpose(-1,-2) @ R2
    cos = rotmat_cosine_angle(R)
    return torch.acos(torch.clamp(cos, -1.0, 1.0))


def unitquat_geodesic_distance(q1, q2):
    """
    Returns the angular distance alpha between rotations represented by **unit** quaternions.
    Based on the equality :math:`min |q_2 \pm q_1| = 2 |sin(alpha/4)|`.

    Args:
        q1, q2 (...x4 tensor, XYZW convention): batch of unit quaternions.
    Returns:
        batch of angles in radians (... tensor).
    """
    return 4.0 * torch.asin(0.5 * torch.min(roma.internal.norm(q2 - q1, dim=-1), roma.internal.norm(q2 + q1, dim=-1)))

def rotvec_geodesic_distance(vec1, vec2):
    """
    Returns the angular distance between rotations represented by rotation vectors.
    (use a conversion to unit quaternions internally).

    Args:
        vec1, vec2 (...x3 tensors): batch of rotation vectors.
    Returns:
        batch of angles in radians (... tensor).
    """
    return unitquat_geodesic_distance(roma.mappings.rotvec_to_unitquat(vec1), roma.mappings.rotvec_to_unitquat(vec2))

def quat_conjugation(quat):
    """
    Returns the conjugation of input batch of quaternions.

    Args:
        quat (...x4 tensor, XYZW convention): batch of quaternions.
    Returns:
        batch of quaternions (...x4 tensor, XYZW convention).
    Note:
        Conjugation of a unit quaternion is equal to its inverse.        
    """
    inv = quat.clone()
    inv[...,:3] *= -1
    return inv

def quat_inverse(quat):
    """
    Returns the inverse of a batch of quaternions.

    Args:
        quat (...x4 tensor, XYZW convention): batch of quaternions.
    Returns:
        batch of quaternions (...x4 tensor, XYZW convention).
    Note:
        - Inverse of null quaternion is undefined.
        - For unit quaternions, consider using conjugation instead.        
    """
    return quat_conjugation(quat) / torch.sum(quat**2, dim=-1, keepdim=True)

def quat_normalize(quat):
    """
    Returns a normalized, unit norm, copy of a batch of quaternions.

    Args:
        quat (...x4 tensor, XYZW convention): batch of quaternions.
    Returns:
        batch of quaternions (...x4 tensor, XYZW convention).        
    """
    return quat / roma.internal.norm(quat, dim=-1, keepdim=True)

def quat_product(p, q):
    """
    Returns the product of two quaternions.

    Args:
        p, q (...x4 tensor, XYZW convention): batch of quaternions.
    Returns:
        batch of quaternions (...x4 tensor, XYZW convention).
    """
    # Adapted from SciPy:
    # https://github.com/scipy/scipy/blob/adc4f4f7bab120ccfab9383aba272954a0a12fb0/scipy/spatial/transform/rotation.py#L153
    # batch_shape = p.shape[:-1]
    # assert q.shape[:-1] == batch_shape, "Incompatible shapes"
    # p = p.reshape(-1, 4)
    # q = q.reshape(-1, 4)
    # product = torch.empty_like(q)
    # product[..., 3] = p[..., 3] * q[..., 3] - torch.sum(p[..., :3] * q[..., :3], axis=-1)
    # product[..., :3] = (p[..., None, 3] * q[..., :3] + q[..., None, 3] * p[..., :3] +
    #                   torch.cross(p[..., :3], q[..., :3], dim=-1))
    
    vector = (p[..., None, 3] * q[..., :3] + q[..., None, 3] * p[..., :3] +
                      torch.cross(p[..., :3], q[..., :3], dim=-1))
    last = p[..., 3] * q[..., 3] - torch.sum(p[..., :3] * q[..., :3], axis=-1)
    return torch.cat((vector, last[...,None]), dim=-1)

def quat_composition(sequence, normalize = False):
    """
    Returns the product of a sequence of quaternions.

    Args:
        sequence (sequence of ...x4 tensors, XYZW convention): sequence of batches of quaternions.
        normalize (bool): it True, normalize the returned quaternion.
    Returns:
        batch of quaternions (...x4 tensor, XYZW convention).
    """
    assert len(sequence) > 1, "Requiring at least two inputs."
    res = sequence[0]
    for q in sequence[1:]:
        res = quat_product(res, q)
    if normalize:
        res = quat_normalize(res)
    return res

def quat_action(q, v, is_normalized=False):
    """
    Rotate a 3D vector :math:`v=(x,y,z)` by a rotation represented by a quaternion `q`.

    Based on the action by conjugation :math:`q,v : q v q^{-1}`, considering the pure quaternion :math:`v=xi + yj +zk` by abuse of notation.

    Args:
        q (...x4 tensor, XYZW convention): batch of quaternions.
        v (...x3 tensor): batch of 3D vectors.
        is_normalized: use True if the input quaternions are already normalized, to avoid unnecessary computations.
    Returns:
        batch of rotated 3D vectors (...x3 tensor).
    Note:
        One should favor rotation matrix representation to rotate multiple vectors by the same rotation efficiently.

    """
    batch_shape = v.shape[:-1]
    iquat = quat_conjugation(q) if is_normalized else quat_inverse(q)
    pure = torch.cat((v, torch.zeros(batch_shape + (1,), dtype=q.dtype, device=q.device)), dim=-1)
    res = quat_product(q, quat_product(pure, iquat))
    return res[...,:3]

def rotvec_inverse(rotvec):
    """
    Returns the inverse of the input rotation expressed using rotation vector representation.

    Args:
        rotvec (...x3 tensor): batch of rotation vectors.
    Returns:
        batch of rotation vectors (...x3 tensor).
    """
    return -rotvec

def rotvec_composition(sequence, normalize = False):
    """
    Returns a rotation vector corresponding to the composition of a sequence of rotations represented by rotation vectors.
    Composition is performed using an intermediary quaternion representation.

    Args:
        sequence (sequence of ...x3 tensors): sequence of batches of rotation vectors.    
        normalize (bool): if True, normalize intermediary representation to compensate for numerical errors.
    """
    assert len(sequence) > 1, "Requiring at least two inputs."
    quats = [roma.rotvec_to_unitquat(rotvec) for rotvec in sequence]
    q = quat_composition(quats, normalize=normalize)
    return roma.unitquat_to_rotvec(q)

def rotmat_inverse(R):
    """
    Returns the inverse of a rotation matrix.

    Args:
        R (...xNxN tensor): batch of rotation matrices.
    Returns:
        batch of inverted rotation matrices (...xNxN tensor).
    Warning:
        The function returns a transposed view of the input, therefore one should be careful with in-place operations.
    """
    return R.transpose(-1, -2)

def rotmat_composition(sequence, normalize = False):
    """
    Returns the product of a sequence of rotation matrices.

    Args:
        sequence (sequence of ...xNxN tensors): sequence of batches of rotation matrices.  
        normalize: if True, apply special Procrustes orthonormalization to compensate for numerical errors.
    Returns:
        batch of rotation matrices (...xNxN tensor).
    """    
    assert len(sequence) > 1, "Requiring at least two inputs."
    result = sequence[0]
    for R in sequence[1:]:
        result = result @ R
    if normalize:
        result = roma.mappings.special_procrustes(result)
    return result

def unitquat_slerp(q0, q1, steps, shortest_arc=True):
    """
    Spherical linear interpolation between two unit quaternions.
    
    Args: 
        q0, q1 (Ax4 tensor): batch of unit quaternions (A may contain multiple dimensions).
        steps (tensor of shape B): interpolation steps, 0.0 corresponding to q0 and 1.0 to q1 (B may contain multiple dimensions).
        shortest_arc (boolean): if True, interpolation will be performed along the shortest arc on SO(3) from `q0` to `q1` or `-q1`.
    Returns: 
        batch of interpolated quaternions (BxAx4 tensor).
    Note:
        When considering quaternions as rotation representations,
        one should keep in mind that spherical interpolation is not necessarily performed along the shortest arc,
        depending on the sign of ``torch.sum(q0*q1,dim=-1)``.

        Behavior is undefined when using ``shortest_arc=False`` with antipodal quaternions.
    """
    # Relative rotation
    rel_q = quat_product(quat_conjugation(q0), q1)
    rel_rotvec = roma.mappings.unitquat_to_rotvec(rel_q, shortest_arc=shortest_arc)
    # Relative rotations to apply
    rel_rotvecs = steps.reshape(steps.shape + (1,) * rel_rotvec.dim()) * rel_rotvec.reshape((1,) * steps.dim() + rel_rotvec.shape)
    rots = roma.mappings.rotvec_to_unitquat(rel_rotvecs.reshape(-1, 3)).reshape(*rel_rotvecs.shape[:-1], 4)
    interpolated_q = quat_product(q0.reshape((1,) * steps.dim() + q0.shape).repeat(steps.shape + (1,) * q0.dim()), rots)
    return interpolated_q

def unitquat_slerp_fast(q0, q1, steps, shortest_arc=True):
    """
    Spherical linear interpolation between two unit quaternions.
    This function requires less computations than :func:`roma.utils.unitquat_slerp`,
    but is **unsuitable for extrapolation (i.e.** ``steps`` **must be within [0,1])**.

    Args: 
        q0, q1 (Ax4 tensor): batch of unit quaternions (A may contain multiple dimensions).
        steps (tensor of shape B): interpolation steps within 0.0 and 1.0, 0.0 corresponding to q0 and 1.0 to q1 (B may contain multiple dimensions).
        shortest_arc (boolean): if True, interpolation will be performed along the shortest arc on SO(3) from `q0` to `q1` or `-q1`.
    Returns: 
        batch of interpolated quaternions (BxAx4 tensor).
    """
    q0, batch_shape = roma.internal.flatten_batch_dims(q0, end_dim=-2)
    q1, batch_shape1 = roma.internal.flatten_batch_dims(q1, end_dim=-2)
    assert batch_shape == batch_shape1
    # omega is the 'angle' between both quaternions
    cos_omega = torch.sum(q0 * q1, dim=-1)
    if shortest_arc:
        # Flip some quaternions to perform shortest arc interpolation.
        q1 = q1.clone()
        q1[cos_omega < 0,:] *= -1
        cos_omega = torch.abs(cos_omega)
    # True when q0 and q1 are close.
    nearby_quaternions = cos_omega > (1.0 - 1e-3)

    cos_omega = cos_omega.reshape((1,) * steps.dim() + (-1,1))
    s = steps.reshape(steps.shape + (1,1))
    # General approach    
    omega = torch.acos(cos_omega)
    alpha = torch.sin((1-s)*omega)
    beta = torch.sin(s*omega)
    # Use linear interpolation for nearby quaternions
    alpha[..., nearby_quaternions, :] = 1 - s
    beta[..., nearby_quaternions, :] = s
    # Interpolation
    q = alpha * q0.reshape((1,) * steps.dim() + q0.shape) + beta * q1.reshape((1,) * steps.dim() + q1.shape)
    # Normalization of the output
    q = quat_normalize(q)
    return q.reshape(steps.shape + batch_shape + (4,))
    
def rotvec_slerp(rotvec0, rotvec1, steps):
    """
    Spherical linear interpolation between two rotation vector representations.

    Args:
        rotvec0, rotvec1 (Ax3 tensor): batch of rotation vectors (A may contain multiple dimensions).
        steps (tensor of shape B):  interpolation steps, 0.0 corresponding to rotvec0 and 1.0 to rotvec1 (B may contain multiple dimensions).
    Returns: 
        batch of interpolated rotation vectors (BxAx3 tensor).
    """
    q0 = roma.mappings.rotvec_to_unitquat(rotvec0)
    q1 = roma.mappings.rotvec_to_unitquat(rotvec1)
    interpolated_q = unitquat_slerp(q0, q1, steps, shortest_arc=True)
    return roma.mappings.unitquat_to_rotvec(interpolated_q)

def rotmat_slerp(R0, R1, steps):
    """
    Spherical linear interpolation between two rotation matrices.

    Args:
        R0, R1 (Ax3x3 tensor): batch of rotation matrices (A may contain multiple dimensions).
        steps (tensor of shape B):  interpolation steps, 0.0 corresponding to R0 and 1.0 to R1 (B may contain multiple dimensions).
    Returns: 
        batch of interpolated rotation matrices (BxAx3x3 tensor).
    """    
    q0 = roma.mappings.rotmat_to_unitquat(R0)
    q1 = roma.mappings.rotmat_to_unitquat(R1)
    interpolated_q = unitquat_slerp(q0, q1, steps, shortest_arc=True)
    return roma.mappings.unitquat_to_rotmat(interpolated_q)

def rigid_vectors_registration(x, y, weights=None, compute_scaling=False):
    """
    Returns the rotation matrix :math:`R` and the optional scaling :math:`s` that best align an input list of vectors :math:`(x_i)_{i=1...n}` to a target list of vectors :math:`(y_i)_{i=1...n}`
    by minimizing the sum of square distance :math:`\sum_i w_i \|s R x_i - y_i\|^2`, where :math:`(w_i)_{i=1...n}` denotes optional positive weights.
    See :func:`~roma.utils.rigid_points_registration` for details.

    Args:
        x (...xNxD tensor): list of N vectors of dimension D.
        y (...xNxD tensor): list of corresponding target vectors.
        weights (None or ...xN tensor): optional list of weights associated to each vector.
    Returns:
        A tuple :math:`(R, s)` consisting of the rotation matrix :math:`R` (...xDxD tensor) and the scaling :math:`s` (... tensor) 
        if :code:`compute_scaling=True`.
        Returns the rotation matrix :math:`R` otherwise.
    """
    if weights is None:
        n = x.shape[-2]
        M = torch.einsum("...ki, ...kj -> ...ij", y, x / n)
    else:
        # Normalize weights
        weights = weights / torch.sum(weights, dim=-1, keepdim=True)
        M = torch.einsum("...k, ...ki, ...kj -> ...ij", weights, y, x)
    if compute_scaling:
        R, DS = roma.special_procrustes(M, return_singular_values=True)
        # Using notations from (Umeyema, IEEE TPAMI 1991).
        DS_trace = torch.sum(DS, dim=-1)
        if weights is None:
            sigma2_x = torch.mean(torch.sum(torch.square(x), dim=-1), dim=-1)
        else:
            sigma2_x = torch.sum(weights * torch.sum(torch.square(x), dim=-1), dim=-1)    
        scale = DS_trace / sigma2_x
        return R, scale
    else:
        R = roma.special_procrustes(M)
        return R

def rigid_points_registration(x, y, weights=None, compute_scaling=False):
    """
    Returns the rigid transformation :math:`(R,t)` and the optional scaling :math:`s` that best align an input list of points :math:`(x_i)_{i=1...n}` to a target list of points :math:`(y_i)_{i=1...n}`,
    by minimizing the sum of square distance :math:`\sum_i w_i \|s R x_i + t - y_i\|^2`, where :math:`(w_i)_{i=1...n}` denotes optional positive weights.
    This is sometimes referred to as the Kabsch/Umeyama algorithm.

    Args:
        x (...xNxD tensor): list of N points of dimension D.
        y (...xNxD tensor): list of corresponding target points.
        weights (None or ...xN tensor): optional list of weights associated to each point.
    Returns:
        a triplet :math:`(R, t, s)` consisting of a rotation matrix :math:`R` (...xDxD tensor), 
        a translation vector :math:`t` (...xD tensor),
        and a scaling :math:`s` (... tensor) if :code:`compute_scaling=True`.
        Returns :math:`(R, t)` otherwise.

    References:
        S. Umeyama, “Least-squares estimation of transformation parameters between two point patterns,” IEEE Transactions on pattern analysis and machine intelligence, vol. 13, no. 4, Art. no. 4, 1991.        

        W. Kabsch, "A solution for the best rotation to relate two sets of vectors". Acta Crystallographica, A32, 1976.
    """
    # Center data
    if weights is None:
        xmean = torch.mean(x, dim=-2, keepdim=True)
        ymean = torch.mean(y, dim=-2, keepdim=True)
    else:
        normalized_weights = weights / torch.sum(weights, dim=-1, keepdim=True)
        xmean = torch.sum(normalized_weights[...,None] * x, dim=-2, keepdim=True)
        ymean = torch.sum(normalized_weights[...,None] * y, dim=-2, keepdim=True)
    xhat = x - xmean
    yhat = y - ymean
    
    # Solve the vectors registration problem
    if compute_scaling:
        R, scale = rigid_vectors_registration(xhat, yhat, weights, compute_scaling=compute_scaling)
        t = (ymean - torch.einsum('...ik, ...jk -> ...ji', scale[...,None, None] * R, xmean)).squeeze(-2)
        return R, t, scale
    else:
        R = rigid_vectors_registration(xhat, yhat, weights, compute_scaling=compute_scaling)
        t = (ymean - torch.einsum('...ik, ...jk -> ...ji', R, xmean)).squeeze(-2)
        return R, t


# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
import torch
import roma
from roma.transforms import *
import unittest
import itertools

class TestTransforms(unittest.TestCase):
    def test_linear_apply(self):
        batch_shape = (10,)
        dtype = torch.float64
        x = torch.randn(batch_shape + (3,), dtype=dtype)
        R = roma.random_rotmat(batch_shape, dtype=dtype)
        t = torch.randn(batch_shape + (3,), dtype=dtype)

        for transform in (Linear, Orthonormal, Rotation, RotationUnitQuat):
            if transform == RotationUnitQuat:
                T = transform(roma.rotmat_to_unitquat(R))
            else:
                T = transform(R)
            Tx = T.apply(x)
            Rx = T.linear_apply(x)
            self.assertTrue(torch.all(torch.isclose(Rx, torch.einsum("...ik, ...k -> ...i", R, x))))
            self.assertTrue(torch.all(torch.isclose(Rx, Tx)))

    def test_apply(self):
        dtype = torch.float64
        batch_shape = (10,)
        x = torch.randn(batch_shape + (3,), dtype=dtype)
        R = roma.random_rotmat(batch_shape, dtype=dtype)
        t = torch.randn(batch_shape + (3,), dtype=dtype) 

        for transform in Affine, Isometry, Rigid, RigidUnitQuat:
            if transform == RigidUnitQuat:
                T = transform(roma.rotmat_to_unitquat(R), t)
            else:
                T = transform(R, t)
            Tx = T.apply(x)
            Rx = T.linear_apply(x)
            self.assertTrue(torch.all(torch.isclose(Rx, torch.einsum("...ik, ...k -> ...i", R, x))))
            self.assertTrue(torch.all(torch.isclose(Tx, Rx + t)))

    def test_composition(self):
        dtype = torch.float64
        batch_shape = (10,5)
        x = torch.randn(batch_shape + (3,), dtype=dtype)
        R1 = roma.random_rotmat(batch_shape, dtype=dtype)
        t1 = torch.randn(batch_shape + (3,), dtype=dtype)
        R2 = roma.random_rotmat(batch_shape, dtype=dtype)
        t2 = torch.randn(batch_shape + (3,), dtype=dtype)

        # Linear transformations
        for transform in (Linear, Orthonormal, Rotation, RotationUnitQuat):
            if transform == RotationUnitQuat:
                T1 = transform(roma.rotmat_to_unitquat(R1))
                T2 = transform(roma.rotmat_to_unitquat(R2))
            else:
                T1 = transform(R1)
                T2 = transform(R2)
            T = T1 @ T2
            self.assertTrue(type(T) == type(T1))

            Tx = T.apply(x)
            Rx = T.linear_apply(x)

            Txbis = T1.apply(T2.apply(x))
            Rxbis = T1.linear_apply(T2.linear_apply(x))
            self.assertTrue(torch.all(torch.isclose(Txbis, Tx)))
            self.assertTrue(torch.all(torch.isclose(Rxbis, Rx)))

        # Affine transformations
        for transform in (Affine, Isometry, Rigid, RigidUnitQuat):
            if transform == RigidUnitQuat:
                T1 = transform(roma.rotmat_to_unitquat(R1), t1)
                T2 = transform(roma.rotmat_to_unitquat(R2), t2)
            else:
                T1 = transform(R1, t1)
                T2 = transform(R2, t2)
            T = T1 @ T2
            self.assertTrue(type(T) == type(T1))

            Tx = T.apply(x)
            Rx = T.linear_apply(x)

            Txbis = T1.apply(T2.apply(x))
            Rxbis = T1.linear_apply(T2.linear_apply(x))
            self.assertTrue(torch.all(torch.isclose(Txbis, Tx)))
            self.assertTrue(torch.all(torch.isclose(Rxbis, Rx)))

    def test_inverse(self):
        batch_shape = (10,)
        dtype = torch.float64
        x = torch.randn(batch_shape + (3,), dtype=dtype)
        R = roma.random_rotmat(batch_shape, dtype=dtype)
        t = torch.randn(batch_shape + (3,), dtype=dtype)

        Ridentity = torch.eye(3, dtype=dtype).reshape([1] * len(batch_shape) + [3,3]).repeat(batch_shape + (1,1))
        tidentity = torch.zeros(batch_shape + (3,), dtype=dtype)

        for transform in (Linear, Orthonormal, Rotation, RotationUnitQuat):
            if transform == RotationUnitQuat:
                T = transform(roma.rotmat_to_unitquat(R))
            else:
                T = transform(R)
            Tinv = T.inverse()
            identity1 = Tinv @ T
            identity2 = T @ Tinv

            if transform == RotationUnitQuat:
                self.assertTrue(torch.all(torch.isclose(roma.unitquat_to_rotmat(identity1.linear), Ridentity)))
                self.assertTrue(torch.all(torch.isclose(roma.unitquat_to_rotmat(identity2.linear), Ridentity)))
            else:
                self.assertTrue(torch.all(torch.isclose(identity1.linear, Ridentity)))
                self.assertTrue(torch.all(torch.isclose(identity2.linear, Ridentity)))

        for transform in (Affine, Isometry, Rigid, RigidUnitQuat):
            if transform == RigidUnitQuat:
                T = transform(roma.rotmat_to_unitquat(R), t)
            else:
                T = transform(R, t)
            Tinv = T.inverse()
            identity1 = Tinv @ T
            identity2 = T @ Tinv

            if transform == RigidUnitQuat:
                self.assertTrue(torch.all(torch.isclose(roma.unitquat_to_rotmat(identity1.linear), Ridentity)))
                self.assertTrue(torch.all(torch.isclose(roma.unitquat_to_rotmat(identity2.linear), Ridentity)))
            else:
                self.assertTrue(torch.all(torch.isclose(identity1.linear, Ridentity)))
                self.assertTrue(torch.all(torch.isclose(identity2.linear, Ridentity)))
            self.assertTrue(torch.all(torch.isclose(identity1.translation, tidentity)))
            self.assertTrue(torch.all(torch.isclose(identity2.translation, tidentity)))

    def test_affine_homogeneous_cast(self):
        batch_shape = (10,)
        for D in range(2,5):
            for Type in (Affine, Isometry, Rigid):
                transformation = Type(torch.randn(batch_shape + (D,D)), torch.randn(batch_shape + (D,))).normalize()
                self.assertTrue(type(transformation) == Type)

                homogeneous = transformation.to_homogeneous()
                self.assertTrue(homogeneous.shape == batch_shape + (D+1, D+1))
                self.assertTrue(torch.all(homogeneous[...,:D,:D] == transformation.linear))
                self.assertTrue(torch.all(homogeneous[...,:D,D] == transformation.translation))
                self.assertTrue(torch.all(homogeneous[...,D,:D] == 0.0))
                self.assertTrue(torch.all(homogeneous[...,D,D] == 1))

                transformation2 = Type.from_homogeneous(homogeneous)
                self.assertTrue(type(transformation2) == Type)
                self.assertTrue(torch.all(transformation.linear == transformation2.linear))
                self.assertTrue(torch.all(transformation.translation == transformation2.translation))

                # Re-use of an existing buffer
                homogeneous_bis = transformation.to_homogeneous(homogeneous)
                self.assertTrue(homogeneous_bis is homogeneous)

    def test_unitquat_homogeneous_cast(self):
        batch_shape = (10,)
        D = 3
        dtype = torch.float64
        Type = RigidUnitQuat
        transformation = Type(torch.randn(batch_shape + (4,), dtype=dtype), torch.randn(batch_shape + (D,), dtype=dtype)).normalize()
        self.assertTrue(type(transformation) == Type)

        homogeneous = transformation.to_homogeneous()
        self.assertTrue(homogeneous.shape == batch_shape + (D+1, D+1))
        self.assertTrue(roma.is_rotation_matrix(homogeneous[...,:D,:D]))
        self.assertTrue(torch.all(roma.unitquat_geodesic_distance(roma.rotmat_to_unitquat(homogeneous[...,:D,:D]), transformation.linear) < 1e-6))
        self.assertTrue(torch.all(homogeneous[...,:D,D] == transformation.translation))
        self.assertTrue(torch.all(homogeneous[...,D,:D] == 0.0))
        self.assertTrue(torch.all(homogeneous[...,D,D] == 1))

        transformation2 = Type.from_homogeneous(homogeneous)
        self.assertTrue(type(transformation2) == Type)
        self.assertTrue(torch.all(roma.unitquat_geodesic_distance(transformation.linear, transformation2.linear) < 1e-6))
        self.assertTrue(torch.all(transformation.translation == transformation2.translation))

        # Re-use of an existing buffer
        homogeneous_bis = transformation.to_homogeneous(homogeneous)
        self.assertTrue(homogeneous_bis is homogeneous)

    def test_orthonormalization(self):
        batch_shape = (10,4)
        D = 5
        dtype = torch.float64

        for _ in range(10):
            raw = torch.randn(batch_shape + (D,D), dtype=dtype)

            ortho1 = Orthonormal(raw).normalize()
            ortho2 = ortho1.normalize()

            self.assertTrue(roma.is_orthonormal_matrix(ortho1.linear))
            self.assertTrue(roma.is_orthonormal_matrix(ortho2.linear))
            self.assertTrue(torch.all(torch.isclose(ortho1.linear, ortho2.linear)))

            translation = torch.randn(batch_shape + (D,), dtype=dtype)
            iso = Isometry(raw, translation).normalize()
            self.assertTrue(torch.all(torch.isclose(iso.linear, ortho1.linear)))
            self.assertTrue(torch.all(torch.isclose(iso.translation, translation)))

    def test_rotation(self):
        batch_shape = (10,4)
        D = 5
        dtype = torch.float64

        for _ in range(10):
            raw = torch.randn(batch_shape + (D,D), dtype=dtype)
            translation = torch.randn(batch_shape + (D,), dtype=dtype)

            rot1 = Rotation(raw).normalize()
            rot2 = rot1.normalize()

            self.assertTrue(roma.is_rotation_matrix(rot1.linear))
            self.assertTrue(roma.is_rotation_matrix(rot2.linear))
            self.assertTrue(torch.all(torch.isclose(rot1.linear, rot2.linear)))

            rigid = Rigid(raw, translation).normalize()
            self.assertTrue(torch.all(torch.isclose(rigid.linear, rot1.linear)))
            self.assertTrue(torch.all(torch.isclose(rigid.translation, translation)))


    def test_rotation_unit_quat(self):
        batch_shape = (10,4)
        D = 3
        dtype = torch.float64

        for _ in range(10):
            raw = torch.randn(batch_shape + (D,D), dtype=dtype)
            translation = torch.randn(batch_shape + (D,), dtype=dtype)

            rot1 = Rotation(raw).normalize()
            quat = RotationUnitQuat(roma.rotmat_to_unitquat(rot1.linear))
            quat1 = quat.normalize()
            self.assertTrue(torch.all(torch.isclose(roma.unitquat_to_rotmat(quat1.linear), rot1.linear)))
            self.assertTrue(torch.all(torch.isclose(quat1.linear, quat.linear)))

            rigidq = RigidUnitQuat(roma.rotmat_to_unitquat(rot1.linear), translation)
            rigidq1 = rigidq.normalize()
            self.assertTrue(torch.all(torch.isclose(roma.unitquat_to_rotmat(rigidq1.linear), rot1.linear)))
            self.assertTrue(torch.all(torch.isclose(rigidq1.translation, translation)))
            self.assertTrue(torch.all(torch.isclose(rigidq.linear, rigidq1.linear)))

    def test_affine_different_dims(self):
        """
        Tests with various input and output spatial dimensions.
        """
        batch_shape = (10,4)
        dtype = torch.float64

        for C in range(2, 5):
            for D in range(2, 5):
                linear = torch.randn(batch_shape + (C, D), dtype=dtype)
                translation = torch.randn(batch_shape + (C,), dtype=dtype)
                roma.Linear(linear)
                x = torch.randn(batch_shape + (D,), dtype=dtype)
                T = roma.Affine(linear, translation)
                Tx = T.apply(x)
                homogeneous = T.to_homogeneous()
                homogeneous2 = roma.Affine.from_homogeneous(homogeneous).to_homogeneous()
                homogeneous3 = roma.Affine.from_homogeneous(homogeneous).to_homogeneous(torch.zeros_like(homogeneous))
                self.assertTrue(homogeneous.shape == batch_shape + (C+1, D+1))
                self.assertTrue(torch.all(torch.isclose(homogeneous, homogeneous2)))
                self.assertTrue(torch.all(torch.isclose(homogeneous, homogeneous3)))
                if C != D:
                    self.assertRaises(AssertionError, lambda : roma.Orthonormal(linear))
                    self.assertRaises(AssertionError, lambda : roma.Rotation(linear))
                    self.assertRaises(AssertionError, lambda : roma.Isometry(linear, translation))
                    self.assertRaises(AssertionError, lambda : roma.Rigid(linear, translation))

    def test_rigid_conversions(self):
        batch_shape = (2,3,6)
        dtype = torch.float64
        rigid = roma.Rigid(roma.random_rotmat(batch_shape, dtype=dtype), torch.zeros(batch_shape + (3,), dtype=dtype))
        rigidunitquat = rigid.to_rigidunitquat()
        rigid2 = rigidunitquat.to_rigid()

        self.assertTrue(torch.all(torch.isclose(rigid.linear, rigid2.linear)))
        self.assertTrue(torch.all(torch.isclose(rigid.translation, rigid2.translation)))

        x = torch.randn(batch_shape + (3,), dtype=dtype)
        x1 = rigid.apply(x)
        x2 = rigidunitquat.apply(x)
        self.assertTrue(torch.all(torch.isclose(x1, x2)))
                    